
DEFAULT_BROM_DIR = "./brom"
DEFAULT_WATCHER_DIR = DEFAULT_BROM_DIR + "/watcher"
DEFAULT_PLOT_DIR = DEFAULT_WATCHER_DIR + "/plots"
DEFAULT_RAW_DATA_DIR = DEFAULT_WATCHER_DIR + "/raw_data"
DEFAULT_BROM_MODELS_DIR = DEFAULT_BROM_DIR + "/models"